from typing import Any, Callable, Generic, TypeVar

from django.db.models import Manager, Model
from django.db.models.query import QuerySet, _QuerySet
from django.db.models.sql import Query
from mypy.nodes import Expression
from typing_extensions import Self

from .meta import CTEColumn, CTEColumns
from .query import CTEQuery

_T_co = TypeVar("_T_co", bound=Model, covariant=True)
_T2_co = TypeVar("_T2_co", bound=Model, covariant=True)
_Row_co = TypeVar("_Row_co", covariant=True)
_Q_co = TypeVar("_Q_co", bound=_QuerySet, covariant=True)
_SelectT = TypeVar("_SelectT", bound=Model | QuerySet | CTE)

def with_cte(*ctes: CTE[Any], select: _SelectT) -> _SelectT: ...

class CTE(Generic[_T_co]):
    query: CTEQuery
    name: str
    col: CTEColumns
    materialized: bool

    def __init__(self, queryset: _QuerySet[_T_co, _Row_co] | None, name: str = ..., materialized: bool = ...) -> None: ...
    def __getstate__(self) -> tuple[CTEQuery | None, str, bool]: ...
    def __setstate__(self, state: tuple[CTEQuery | None, str, bool]) -> None: ...
    @classmethod
    def recursive(
        cls,
        make_cte_queryset: Callable[[Self], _QuerySet[_T2_co, _Row_co]],
        name: str = ...,
        materialized: bool = ...,
    ) -> Self: ...
    def join(
        self,
        *filter_q: type[_T2_co | _T_co] | _QuerySet[_T_co, _Row_co],
        **filter_kw: CTEColumn,
    ) -> CTEQuerySet[_T_co]: ...
    def queryset(self) -> CTEQuerySet[_T_co]: ...
    def _resolve_ref(self, name: str) -> Expression: ...
    def resolve_expression(
        self,
        query: Any | None = None,
        allow_joins: bool = True,
        reuse: set[str] | None = None,
        summarize: bool = False,
        for_save: bool = False,
    ) -> Self: ...

class CTEQuerySet(QuerySet[_T_co]):
    def __init__(
        self,
        model: type[Model] | None = ...,
        query: Query | None = ...,
        using: str | None = ...,
        hints: dict[str, Model] | None = ...,
    ) -> None: ...
    @classmethod
    def as_manager(cls) -> CTEManager[_T_co]: ...

class ManagedCTEQuerySet(CTEQuerySet[_T_co], Generic[_T_co, _Q_co]):
    def __init__(
        self,
        model: type[Model] | None = ...,
        query: Query | None = ...,
        using: str | None = ...,
        hints: dict[str, Model] | None = ...,
    ) -> None: ...
    @classmethod
    def as_manager(cls) -> CTEManager[_T_co]: ...

class CTEManager(Manager[_T_co]):
    @classmethod
    def from_queryset(cls, queryset_class: type[_QuerySet[_T_co, _T_co]], class_name: str | None = ...) -> type[Self]: ...



from collections.abc import Mapping
from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Field
from django.db.models.expressions import Expression
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import Self

from .cte import CTE

class CTEColumns:
    def __init__(self, cte: CTE) -> None: ...
    def __getattr__(self, name: str) -> CTEColumn: ...

class CTEColumn(Expression):
    table_alias: str
    name: str
    alias: str | None
    _cte: CTE
    _output_field: Field | None

    def __init__(self, cte: CTE, name: str, output_field: Field | None = ...) -> None: ...
    @property
    def _ref(self) -> Expression: ...
    @property
    def target(self) -> Field: ...
    @property
    def output_field(self) -> Field: ...
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def relabeled_clone(self, relabels: Mapping[str, str]) -> CTEColumn: ...

class CTEColumnRef(Expression):
    name: str
    cte_name: str
    output_field: Field
    _alias: str | None

    def __init__(self, name: str, cte_name: str, output_field: Field) -> None: ...
    def resolve_expression(
        self,
        query: Any = ...,
        allow_joins: bool = ...,
        reuse: set[str] | None = ...,
        summarize: bool = ...,
        for_save: bool = ...,
    ) -> CTEColumnRef: ...
    def relabeled_clone(self, change_map: Mapping[str, str]) -> Self: ...
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...



from typing import Literal

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from django.db.models.sql.where import WhereNode
from typing_extensions import Self

class QJoin:
    filtered_relation: None
    parent_alias: str
    table_name: str
    table_alias: str
    on_clause: WhereNode
    join_type: Literal["INNER JOIN", "LEFT OUTER JOIN"]
    nullable: bool

    def __init__(
        self,
        parent_alias: str,
        table_name: str,
        table_alias: str,
        on_clause: WhereNode,
        join_type: Literal["INNER JOIN", "LEFT OUTER JOIN"] = ...,
        nullable: bool | None = ...,
    ) -> None: ...
    @property
    def identity(self) -> type[QJoin]: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def equals(self, other: object) -> bool: ...
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def relabeled_clone(self, change_map: dict[str | None, str]) -> Self: ...

    class join_field:  # noqa: N801
        class related_model:  # noqa: N801
            class _meta:  # noqa: N801
                local_concrete_fields: tuple



from collections.abc import Mapping, Sequence
from typing import Any, TypeVar

from django.db.models import Field, Model, QuerySet

class RawCteSqlModel(Model): ...

_T_RawCteSqlModel = TypeVar("_T_RawCteSqlModel", bound=RawCteSqlModel)

def raw_cte_sql(sql: str, params: Sequence[Any], refs: Mapping[str, Field]) -> QuerySet[_T_RawCteSqlModel]: ...



from typing import Any, Callable

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Model
from django.db.models.sql import DeleteQuery, Query, UpdateQuery
from django.db.models.sql.compiler import (
    SQLUpdateCompiler,
    _AsSqlType,
)
from django.db.models.sql.where import WhereNode

from .cte import CTE

class CTEQuery(Query):
    _jit_mixin_prefix: str
    _with_ctes: tuple
    def __init__(self, model: type[Model] | None, where: type[WhereNode] = ..., alias_cols: bool = ...) -> None: ...
    @property
    def combined_queries(self) -> tuple: ...
    @combined_queries.setter
    def combined_queries(self, queries: tuple) -> None: ...
    def resolve_expression(self, *args: Any, **kwargs: Any) -> CTEQuery: ...
    def get_compiler(self, *args: Any, **kwargs: Any) -> CTECompiler: ...
    def chain(self, klass: type[Query] | None = ...) -> CTEQuery: ...

def generate_cte_sql(connection: BaseDatabaseWrapper, query: Query, as_sql: Callable) -> tuple[str, tuple]: ...
def get_cte_query_template(cte: CTE) -> str: ...
def _ignore_with_col_aliases(cte_query: Query) -> None: ...

class CTECompiler:
    _jit_mixin_prefix: str
    def as_sql(self, *args: Any, **kwargs: Any) -> _AsSqlType: ...

class NoAliasQuery:
    _jit_mixin_prefix: str
    def get_compiler(self, *args: Any, **kwargs: Any) -> NoAliasCompiler: ...

class NoAliasCompiler:
    _jit_mixin_prefix: str
    def get_select(self, *, with_col_aliases: bool = False, **kw: Any) -> Any: ...

class CTEUpdateQuery(UpdateQuery, CTEQuery): ...
class CTEDeleteQuery(DeleteQuery, CTEQuery): ...

QUERY_TYPES: dict[type[Query], type[CTEQuery]]

COMPILER_TYPES: dict[type[CTEQuery], type[SQLUpdateCompiler]]
